# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::Approp.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Appropriateness for Data::FStructures
#
###############################################################

package Data::FStructure::Approp;
# System modules
use Carp;
use Exporter;
# 3rd party exstensions
# user extension modules
use Math::PartialOrder qw($TYPE_TOP);
use Data::FStructure::Base;
@ISA       = qw();
@EXPORT_OK = qw($APPR_KEY $APPR_WANT_USER_HOOKS);
%EXPORT_TAGS = (
		'apprvars' => [qw($APPR_KEY $APPR_WANT_USER_HOOKS)]
	       );

###############################################################
# Package-global variables
###############################################################
our $VERSION = 0.08;

# where to store appr-information in the hierarchy
our $APPR_KEY = '__APPR__';
our $APPR_WANT_USER_HOOKS = 1;

###############################################################
# Hierarchy Extensions
###############################################################
package Math::PartialOrder;
Data::FStructure::Approp->import qw(:apprvars);

#--------------------------------------------------------------
# Appropriateness Basics: the 'appr' attribute
#
# + Uses _hattributes key $APPR_KEY
# + points to a hashref:
#   { LiteralFeature1 => { FromType1a => ApprVal1a, ...}, ...}
#--------------------------------------------------------------

#--------------------------------------------------------------
# Low-Level Hierarchy-Appr-Methods
#--------------------------------------------------------------
# $apprhash = $hi->appr();
# --> always returns a hashref -- automagic creation!
sub appr ($) {
  return
    (defined($_[0]->_hattributes->{$APPR_KEY})
     ? $_[0]->_hattributes->{$APPR_KEY}
     : ($_[0]->_hattributes->{$APPR_KEY} = {}));
}

#--------------------------------------------------------------
# Mid-Level Hierarchy-Appr-Methods : has_appr()
#--------------------------------------------------------------

# $bool = $hi->has_appr();
# $bool = $hi->has_appr($ftr)
# $bool = $hi->has_appr($ftr,$type)
# --> test whether any appr info is stored -- no autocreate!
sub has_appr ($;$$) {
  my ($appr);
  return '' unless (defined($appr = shift->appr));
  return
    (defined($_[1])
     ? (exists($appr->{$_[1]})
	? (defined($_[2])
	   ? (exists($appr->{$_[1]}{$_[2]}) ? 1 : 0)
	   : 1)
	: 0)
     : 1);
}

#--------------------------------------------------------------
# Mid-Level Hierarchy-Appr-Methods : get/set/clear
#--------------------------------------------------------------

# \%apprhash = $h->set_appr(\%apprhash)
sub set_appr ($$) { %{$_[0]->appr} = %{$_[1]}; return $_[0]->appr; }


#   $t2v = $h->get_appr_byfeat($feat)
#     $v = $h->get_appr_byfeat($feat,$type)
# ---> returns $TYPE_TOP if none found
sub get_appr_byfeat ($$;$) {
  # be sane... please
  unless (defined($_[1])) {
    carp("Warning: `undef' features not supported in get_appr_byfeat()");
    return $TYPE_TOP;
  }
  my $appr = shift->appr();
  my ($f,$t) = @_;

  # lookup feature-hash
  return
    (defined($t)
     && exists($appr->{$f}) && defined($appr->{$f})
     && exists($appr->{$f}{$t})
     ? $appr->{$f}{$t}
     : $TYPE_TOP);
}

#   $t2v = $h->get_appr_bytype($type)
#     $v = $h->get_appr_bytype($tye,$feat)
# ---> returns empty hash if none found
sub get_appr_bytype ($$;$) {
  return $_[0]->get_appr_byfeat($_[2],$_[1]) if (exists($_[2]));

  my ($appr,$t) = (shift->appr,shift);
  my %tappr = ();
  foreach (keys(%$appr)) {
    $tappr{$_} = $appr->{$_}{$t} if (exists($appr->{$_}{$t}));
  }
  return \%tappr;
}


# $h->set_appr_byfeat($feat,$type,$val);
# $h->set_appr_bytype($type,$feat,$val);
# ---> returns $val
sub set_appr_byfeat ($$$$) { return $_[0]->set_appr_bytype($_[2],$_[1],$_[3]); }
sub set_appr_bytype ($$$$) {
  my $appr = shift->appr();
  my ($type,$feat,$val) = @_;

  # be sane... please
  unless (defined($feat)) {
    carp("Warning: `undef' features not supported in set_appr_by() methods");
    return undef;
  }
  unless (defined($type)) {
    carp("Warning: `undef' source types not supported in set_appr_by() methods");
    return undef;
  }

  # actually do the assignment
  $appr->{$feat}{$type} = $val;
}

# undef = $h->remove_appr_byfeat($ftr);
# undef = $h->remove_appr_byfeat($ftr,$type);
sub remove_appr_byfeat ($$;$) {
  return undef unless (defined($_[1]));
  my $appr = $_[0]->appr;
  if (exists($_[2]) && defined($_[2])) {
    if (exists($appr->{$_[1]})) {
      delete($appr->{$_[1]}{$_[2]});
      delete($appr->{$_[1]}) unless (%{$appr->{$_[1]}});
    }
  } else {
    delete($appr->{$_[1]});
  }
}

# undef = $h->remove_appr_bytype($from);
# undef = $h->remove_appr_byfeat($from,$ftr);
sub remove_appr_bytype ($$;$) {
  return $_[0]->remove_appr_byfeat($_[2],$_[1]) if (exists($_[2]));
  my ($h,$t) = @_;
  return undef unless (defined($t));
  my $appr = $h->appr;
  foreach (keys(%$appr)) {
    delete($appr->{$_}{$t}) if (exists($appr->{$_}{$t}));
  }
}



# undef = $h->merge_appr(\%apprhash)
sub merge_appr {
  my $appr = shift->appr;
  %$appr = (%$appr, %{$_[1]});
}

# undef = $h->clear_appr()
sub clear_appr { %{$_[0]->appr} = (); }


#--------------------------------------------------------------
# Mid-Level Approp Methods : $h->get_approp_byfeat()
#--------------------------------------------------------------

# @vals = $h->get_approp_byfeat($feat,$type)
sub get_approp_byfeat ($$$) {
  my ($h,$ftr,$from) = @_;

  # no features appropriate for `undef'
  return qw() unless (defined($from));

  # be sane... please
  unless (defined($ftr)) {
    carp("Warning: `undef' feature not supported in _get_approp_tf()");
    return qw();
  }

  # get the appr-hash: we'll be needing it
  my $appr = $h->appr;

  # user hook
  if ($APPR_WANT_USER_HOOKS && UNIVERSAL::can($from,'get_approp_byfeat')) {
    return $from->get_approp_byfeat($ftr);
  }

  # easy answer
  return ($appr->{$ftr}{$from}) if
    (exists($appr->{$ftr}) && exists($appr->{$ftr}{$from}));

  # we must needs iterate...
  return
    map {
      $appr->{$ftr}{$_}
    } ($h->max_subsuming($from, keys(%{$appr->{$ftr}})));
}

#--------------------------------------------------------------
# Mid-Level Approp Methods : $h->get_approp_bytype($base)
#--------------------------------------------------------------

# \%ftr2allvals = $h->get_approp_bytype($base_type)
#         @vals = $h->get_approp_bytype($base_type,$ftr)
sub get_approp_bytype ($$;$) {
  return $_[0]->get_approp_byfeat($_[2],$_[1]) if (defined($_[2]));
  my $h = shift;
  my $base = shift;

  # no features appropriate for `undef'
  return {} unless (defined($base));

  my $appr = $h->appr;
  my $maxappr = {};
  my (@maxvals);
  foreach (keys(%$appr)) {
    @maxvals = $h->get_approp_byfeat($_,$base);
    $maxappr->{$_} = [@maxvals] if (@maxvals);
  }
  return $maxappr;
}



#--------------------------------------------------------------
# High-Level Approp Methods & Hooks : is_approp
#--------------------------------------------------------------

# $bool = $h->is_approp($fromtype,\%feats2values)
sub is_approp {
  my ($ans);
  return
    (# easy answers
     defined($ans = _is_approp_trivial($_[1],$_[2])) ? $ans
     : (# user-hooks
	($APPR_WANT_USER_HOOKS && defined($ans = _is_approp_user($_[1],$_[2])) ? $ans
	 : (# ... do the lookup
	    $_[0]->_is_approp($_[1],$_[2])))));
}

# $bool_or_undef = _is_approp_trivial($from,\%f2v)
# --> checks for empty \%f2v
sub _is_approp_trivial { return (%{$_[1]} ? undef : 1); }

# $bool_or_undef = _is_approp_user($from,\%f2v)
# --> calls $from->is_approp(\%f2v) if available
sub _is_approp_user {
  return
    (UNIVERSAL::can($_[0],'is_approp')
     ? $_[0]->is_approp($_[1])
     : undef);
}

# $bool = $h->_is_approp($fromtype,\%feats2values)
# ---> hierarchy appr-delegated approp-lookup
sub _is_approp {
  my ($h,$from,$f2v) = @_;
  return 1 unless (%$f2v);
  my ($f,@maxvals);
  foreach $f (keys(%$f2v)) {
    foreach (@maxvals = $h->get_approp_byfeat($f,$from)) {
      return '' unless ($h->subsumes($_,$f2v->{$f}));
    }
    return '' unless (@maxvals); # nothing approp
  }
  return 1;
}



#--------------------------------------------------------------
# High-Level Approp Methods & Hooks : is_totally_approp
#--------------------------------------------------------------

# $bool = $h->is_totally_approp($fromtype,$f2v);
sub is_totally_approp {
  my ($ans);
  return
    (($APPR_WANT_USER_HOOKS && defined($ans = _is_totally_approp_user($_[1],$_[2])))
     ? $ans
     : $_[0]->_is_totally_approp($_[1],$_[2]));
}


# $bool_or_undef = _is_totally_approp_user($from,\%f2v)
# --> calls $from->is_totally_approp(\%f2v) if available
sub _is_totally_approp_user {
  if (UNIVERSAL::can($_[0],'is_totally_approp')) {
    return $_[1]->is_totally_approp($_[1]);
  }
  elsif (UNIVERSAL::can($_[0],'is_approp')) {
    foreach (keys(%{$_[1]})) {
      # just check off the user-defined keys
      delete($_[1]->{$_}) if ($_[0]->is_approp($_,$_[1]->{$_}));
    }
  }
  return undef;
}

# $bool = _is_totally_approp($from,\%f2v)
sub _is_totally_approp {
  my ($h,$from,$f2v) = @_;

  # no features appropriate for undef
  unless (defined($from)) { return (%$f2v) ? 0 : 1; }

  # get the full approp-hash : we'll need it
  my $approp = $h->get_approp_bytype($from);

  # pathological case
  return '' if (%$f2v && !%$approp);

  my ($f,$v);
  foreach $f (keys(%$f2v)) {
    # can $from take $f at all?
    return '' unless (exists($approp->{$f}) && @{$approp->{$f}});
    foreach $v (@{$approp->{$f}}) {
      # is it well-typed?
      return '' unless
	((defined($v) && $v eq $TYPE_TOP)
	 || $h->subsumes($v, $f2v->{$f}));
    }
    # ... then mark the feature as present
    delete($approp->{$f});
  }
  # remove TOP (negated approp-override)
  delete(@{$approp}{grep { $approp->{$_}[0] eq $TYPE_TOP } keys(%$approp)});

  # did we find all the features?
  return (%$approp ? 0 : 1);
}

#--------------------------------------------------------------
# High-Level Approp Methods & Hooks : get_approp_det()
#--------------------------------------------------------------

# $type = $h->get_approp_det($base,$ftr);
# $f2v = $h->get_approp_det($base);
sub get_approp_det ($$;$) {
  if (exists($_[2])) {
    # ... we have a feature
    return $_[0]->njoin($_[0]->get_approp_byfeat($_[2],$_[1]));
  }
  # no feature, just type
  my $approp = $_[0]->get_approp_bytype($_[1]);
  @$approp{keys(%$approp)} = map { $_[0]->njoin(@$_) } values(%$approp);
  return $approp;
}




###############################################################
# FStructure Extensions
###############################################################
package Data::FStructure::Base;


#--------------------------------------------------------------
# Well-typedness
#--------------------------------------------------------------

# $bool = $fs->is_well_typed()
# $bool = $fs->is_well_typed($hi)
#  + alias: is_wt();
*is_wt = \&is_well_typed;
sub is_well_typed {
  my $fs = shift;
  my ($h);
  unless (defined($h = shift || $fs->hierarchy)) {
    carp("Warning: method `is_well_typed()' requires a hierarchy");
    return undef;
  }
  my ($q,%valhash);
  foreach $q ($fs->nodes) {
    %valhash = map { $_ => $fs->theta($fs->delta($q,$_)) } $fs->features($q);
    return '' unless (!%valhash || $h->is_approp($fs->theta($q),\%valhash));
  }
  return 1;
}


#--------------------------------------------------------------
# Total well-typedness
#--------------------------------------------------------------

# $bool = $fs->is_totally_well_typed()
# $bool = $fs->is_totally_well_typed($hi)
#  + alias: is_twt();
*is_twt = \&is_totally_well_typed;
sub is_totally_well_typed {
  my $fs = shift;
  my ($h);
  unless (defined($h = shift || $fs->hierarchy)) {
    carp("Warning: method `is_totally_well_typed()' requires a hierarchy");
    return undef;
  }
  my ($q,%valhash);
  foreach $q ($fs->nodes) {
    %valhash = map { $_ => $fs->theta($fs->delta($q,$_)) } $fs->features($q);
    return '' unless ($h->is_totally_approp($fs->theta($q),\%valhash));
  }
  return 1;
}


#--------------------------------------------------------------
# Type-Inference
#--------------------------------------------------------------
# $fs_or_undef = $fs->make_well_typed();
# $fs_or_undef = $fs->make_well_typed($hi);
# Alias: typinf, make_wt
*typinf = \&make_well_typed;
*make_wt = \&make_well_typed;
sub make_well_typed {
  my $fs = shift;

  carp("Warning: method make_well_typed() is not yet implemented!");
  return $fs;

  my ($h);
  unless (defined($h = shift || $fs->hierarchy)) {
    carp("Warning: method `make_well_typed()' requires a hierarchy");
    return undef;
  }
  my ($step,$q,%valhash,$type);
  my $changed = 1;
  for ($step = scalar($fs->nodes); $changed && ($step >= 0); --$step) {
    foreach $q ($fs->nodes) {
      %valhash = map { $_ => $fs->theta($fs->delta($q,$_)) } $fs->features($q);
      next if (!%valhash);
      $type = $h->make_approp($fs->theta($q),\%valhash);
      if ($type != $fs->theta($q)) {
	$changed = 1;
	$fs->theta($q,$type);
      }
    }
  }
  return undef if ($step <= 0);
  return $fs;
}

#--------------------------------------------------------------
# Total Type-Inference
#--------------------------------------------------------------
# $fs_or_undef = $fs->make_totally_well_typed();
# $fs_or_undef = $fs->make_totally_well_typed($hi);
# Alias: make_twt, fill
# NOT YET IMPLEMENTED
*typfill = \&make_totally_well_typed;
*make_twt = \&make_totally_well_typed;
sub make_totally_well_typed {
  my $fs = shift;
  carp("Warning: method make_totally_well_typed() is not yet implemented!");
  return $fs;
}

1;
__END__

