# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::Array.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Feature-structures for perl: FS arrays
#
###############################################################

package Data::FStructure::Array;
# System modules
# 3rd party exstensions
# user extension modules
use Data::FStructure::Std;
use Data::FStructure::theta;
use Data::FStructure::arydelta qw(:arrayvars);
@ISA = qw(Data::FStructure::Base);
#@EXPORT    = qw();
#@EXPORT_OK = (
#	     );
#%EXPORT_TAGS =
#  (
#  );

###############################################################
# Package-global variables
###############################################################
our $VERSION = 0.02;

###############################################################
# Initialization:
# $fs->initialize(\%args)
# + new keys:
#    elts => [ $fs1, ..., $fsN ]
###############################################################
sub initialize {
  my ($fs,$args) = @_;
  $fs->{theta} = Data::FStructure::theta->new()
    unless (defined($fs->{theta}));
  $fs->{delta} = Data::FStructure::arydelta->new()
    unless (defined($fs->{delta}));
  $fs->elts($args->{elts}) if (defined($args->{elts}));
  return $fs;
}

###############################################################
# Accessors/manipulators
###############################################################

# \@node_array = $fs->node_array();
# \@node_array = $fs->node_array($from);
# alias: nodeary();
*node_array = \&nodeary;
sub nodeary {
  return
    (defined($_[1])
     ? ($_[0]->{delta}->elts($_[1]))
     : ($_[0]->{delta}->elts($_[0]->{q0})));
}

# \@subarray = $fs->elt_array();
# \@subarray = $fs->elt_array($from);
# alias: elts()
*elt_array = \&elts;
sub elts {
  my ($nodes);
  return
    (exists($_[1])
     ? (defined($nodes = $_[0]->nodeary($_[1]))
	? [map { $_[0]->substructure($_) } @$nodes]
	: [])
     : (defined($nodes = $_[0]->nodeary($_[0]->{q0}))
	? [map { $_[0]->substructure($_) } @$nodes]
	: []));
}


# set perldata vars if available --- now done in perldata.pm
#if (defined(%Data::FStructure::perldata::SPIT_TYPE2REF) &&
#    defined(%Data::FStructure::perldata::SPIT_LINKUS))
#  {
#    $Data::FStructure::perldata::SPIT_TYPE2REF{DEFAULT} = \&_type2ref_fsarray;
#    $Data::FStructure::perldata::SPIT_LINKUS{DEFAULT} = \&_linkus_fsarray;
#  }

###############################################################
# gobble & spit
###############################################################



###############################################################
# Everything else ought to be inherited...
###############################################################

1;
__END__

# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::Array - Feature structure arrays for perl.

=head1 SYNOPSIS

  use Data::FStructure::Array;

  #
  # Construction
  #
  $fs = Data::FStructure::Array->new();     # new FStructure, may contain arrays

  #
  # Get/set array indices
  #
  $node = $fs->delta($node,$i);            # numeric features are array-indices
  $to = $fs->delta($node,$i,$val);         # ... also for lookup

  #
  # Get perl arrays
  #
  $ary = $fs->node_array($from);           # get an array-ref of nodes
  $elts = $fs->elt_array($from);           # ... or of substructures

  #
  # perldata interface
  #
  $fs->_gobble_array($aryref,              # gobble arrays safely
		     \%val2nodes,
		     \@stack,
		     \%opts);

  _type2ref_fsarray($fs,$type,$q);         # spit arrays safely
  _linkus_fsarray($fs,                     # ... and link them
		  $qfrom,$f,$qto,
		  \%node2ref);


=head1 REQUIREMENTS

Carp,
Data::FStructure::Base,
Data::FStructure::theta,
Data::FStructure::arydelta.

=cut

#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

C<Data::FStructure::Array> is a base class for FStructure objects
which may contain arrays.  It uses perl array-references as the
underlying representation of arrays, and maintains compatibility
with C<Data::FStructure::Std> and C<Data::FStructure::Base> by
means of the C<Data::FStructure::arydelta> class.

=cut

#--------------------------------------------------------------
# Variables
#--------------------------------------------------------------
=pod

=head1 PACKAGE VARIABLES

C<Data::FStructure::Array> does not define any package
variables, but it does import the variable
C<$ARRAY_FTR> from C<Data::FStructure::arydelta>.
See L<Data::FStructure::arydelta> for details on
the pseudo-feature C<$ARRAY_FTR>.

=cut

#--------------------------------------------------------------
# Methods
#--------------------------------------------------------------
=pod

=head1 METHODS

=cut

#--------------------------------------------------------------
# Construction
#--------------------------------------------------------------
=pod

=head2 Initialization

=over 4

=item * C<intialize(\%args)>

C<Data::FStructure::Array> (and its descendants such as
C<Data::FStructure>) recognize the key argument C<elts>
in C<\%args>.  If specified, C<elts> should be a reference
to an array which will be read in to initialize the
FStructure's contents.  Default is no array at all.

=back

=cut


#'
#--------------------------------------------------------------
# array methods
#--------------------------------------------------------------
=pod

=head2 Array-related Methods

=over 4

=item * C<delta($node,$i)>, C<delta($node,$i,$val)>

Get/set the array-element at the numeric index $i of
the array associated with C<$node>.  The array will
be created for the 3-argument form of C<delta()> if
it does not already exist.

Note that the ability to get and set array-indices
does not in any way affect the ability of
C<Data::FStructure::Array> objects to represent
or store symbolic features.

=item * C<node_array()>, C<node_array($from)>

Alias: C<nodeary()>

Returns the array-reference associated with the node
C<$from>, which defaults to the root of the FStructure,
or C<undef> if no array is currently associated with
C<$from>.  Contents of the array are node-names.

=item * C<elt_array()>, C<elts($from)>

Alias: C<elts()>

Like C<node_array()>, but always returns a new array-reference
whose contents are substructures (as created with the C<substructure()>
method), rather than node-names.

=back

=cut

#--------------------------------------------------------------
# perldata
#--------------------------------------------------------------
=pod

=head2 perldata Interface

=over 4

=item * C<_gobble_array()>

This method will be called by the C<gobble()> method
to ingest perl arrays when they are read in.
See L<perldata> for details.

=item * C<_type2ref_fsarray()>, C<_linkus_fsarray()>

If C<Data::FStructure::perldata> is present when
C<Data::FStructure::Array> is loaded, these method
will be assigned as the default processors
for C<spit()>.  It maps only nodes with associated
arrays which contain no additional symbolic features
to perl array-references -- everything else is handled
by the default C<perldata> functions.

=back

=cut


#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 BUGS

It's missing some high-level calls and a tied interface.

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Data::FStructure(3pm).
Data::FStructure::Std(3pm).
Data::FStructure::List(3pm).
Data::FStructure::theta(3pm).
Data::FStructure::arydelta(3pm).
