# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::Tied.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Tied access to Data::FStructure objects
#              as recursive hashes keyed by features
#
###############################################################

package Data::FStructure::TiedHash;
# System modules
#use Carp;
use Tie::Hash;
# 3rd party extensions
# user extensions
use Data::FStructure;
@ISA = qw(Data::FStructure Tie::Hash);

$VERSION = 0.03;

our %GOBOPTS = (
		reuse_paths => 1,
		untie_hashes => 1,
		untie_arrays => 1
	       );

###############################################################
#   + Object-structure:
#     {
#      --- inherited ---
#      hierarchy       => Math::PartialOrder
#      Q (nodes)       => { id1 => node1, ... } (blessed as Set::Hashed)
#      q0 (root)       => Scalar node
#      qcounter        => Scalar integer (counts nodes)
#      master          => Scalar-ref: root of largest embedding FStructure
#      #theta (types)   => qd::fs::theta
#      #delta (ftrs)    => qd::fs::delta
#      #pathsep (opt)   => Scalar path-separator
#      #pathre (opt)    => Scalar pathsep-regexp
#      --- new ---
#      #val2node        => {$val1 => $node1, ...}
#      featkeys        => [ftr1,ftr2,...]   # ALWAYS NEW! (used for iterators)
#     }
###############################################################

# \%opts = $fs->_gobble_options
sub _gobble_options { return \%GOBOPTS; }

###############################################################
# Low-level functions
###############################################################

# $fs2 = $fs1->substructure($node)
sub substructure {
  my $fs = shift;
  $fs = $fs->SUPER::substructure(@_);
  # always store new 'savekeys'
  $fs->{featkeys} = [];
  return $fs;
}

# $fs = $fs->clear()
sub clear {
  #%{$_[0]->{val2node}} = ();
  @{$_[0]->{featkeys}} = qw();
  return $_[0]->SUPER::clear();
}

###############################################################
# Conventions
#   + keys are single features
#   + values are *references* to hashes tied to the
#     substructures at the key-feature
###############################################################


# TIEHASH($ClassName,@LIST) => $object
#  @LIST = ( \%args )
#  + additional \%args
#     fs => $subfs
#     --> tie to $fs
sub TIEHASH {
  my ($class,$args) = @_;
  $args = {} unless (defined($args));
  my ($this);

  if (exists($args->{fs}) && UNIVERSAL::isa($args->{fs}, 'Data::FStructure::Base')) {
    $this = {%{$args->{fs}}};
  } else {
    $this = $class->SUPER::new($args);
  }
  bless $this, $class;

  # must have root node!
  $this->root($this->_gensym_node()) unless (defined($this->root));

  #$this->{val2node} = {} unless (exists($this->{val2node}));
  $this->{featkeys} = []; # featkeys() always new!
  $this->{gobopts} = \%GOBOPTS;

  return $this;
}

# STORE($this,$key,$value) => \%tied_to_value_substruc
sub STORE {
  my $val = $_[2];
  my ($sub,%h);

  # set the value
  $_[0]->gobble($val,
		%{$_[0]->_gobble_options},
		#val2node=>$_[0]->{val2node},
		at=>$_[1]);
  $sub = $_[0]->substructure($_[0]->delta($_[0]->root,$_[1]));

  # something went wrong
  return undef unless (defined($sub));

  # return a new tied hash
  tie(%h, ref($_[0]), { fs => $sub });
  return \%h;
}

# FETCH($this,$key) => \%tied_to_key
sub FETCH {
  # automagically create paths on fetch...
  if (!defined($_[0]->delta($_[0]->root, $_[1]))) {
    my $node = $_[0]->add_node;
    $_[0]->delta($_[0]->root,$_[1],$node);
  }
  my (%h);
  tie(%h,
      ref($_[0]),
      { fs => $_[0]->substructure($_[0]->delta($_[0]->root, $_[1])) });
  return \%h;
}

# FIRSTKEY($this) => ($key,$val)
sub FIRSTKEY {
  # magically reset any open 'each' iterators
  @{$_[0]->{featkeys}} = $_[0]->features($_[0]->root);
  return pop(@{$_[0]->{featkeys}});
}

# NEXTKEY($this,$lastkey) => ($key,$val)
sub NEXTKEY { return pop(@{$_[0]->{featkeys}}); }

# EXISTS($this,$key) => $bool
sub EXISTS {
  return defined($_[0]->delta($_[0]->root, $_[1]));
}

# DELETE($this,$key)
sub DELETE {
  return undef unless (EXISTS(@_));
  $_[0]->{delta}->remove_transit($_[0]->root, $_[1]);
}

# CLEAR($this)
sub CLEAR { $_[0]->clear; }



###############################################################
# in case anyone tried it...
###############################################################

package Data::FStructure::Tied;
@ISA = qw(Data::FStructure::TiedHash);
our $VERSION = $Data::FStructure::TiedHash::VERSION;

1;
__END__


#--------------------------------------------------------------
# pod Header
#--------------------------------------------------------------
=pod

=head1 NAME

Data::FStructure::TiedHash --
Tied interface to C<Data::FStructure> feature-structures
as recursive hashes.

=head1 SYNOPSIS

  use Data::FStructure::Tied;

  my $h = {};                         # because hashrefs are prettier
  tie %$h, Data::FStructure::TiedHash; # tie the hash to a feature-structure

  $h->{foo} = 'bonk';                 # set feature-value
  $h->{bar}{baz}{bip} = 'blap';       # set path-value
  $hfoo = $h->{foo};                  # get feature-value as tied hashref
  $hnew = $h->{new};                  # $h->{new} is automagically created
  $h->{new} = $h->{foo};              # do structure-sharing

  delete($h->{new});                  # eliminate feature-transition

  @features = keys(%$h);              # get immediate features
  @hashes = values(%$h);              # get direct substructure-hashes
  while (($f,$v) = each(%$h)) {...}   # iterate over direct-substructure hashes

  my $fs = tied(%$h);                 # recover underlying FStructure object

  $fs->remove_unreachable;            # housekeeping
  print $fs->as_pretty, "\n";         # ... or any other FStructure method

=cut


#--------------------------------------------------------------
# Requirements
#--------------------------------------------------------------
=pod

=head1 REQUIREMENTS

Data::FStructure,
Data::FStructure::Safe,
Tie::Hash.

=cut


#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

The C<Data::FStructure::TiedHash> module provides a tied-hash
interface to the C<Data::FStructure> feature-structure class.
Features appear as keys of the hash,
and hash-values are themselves references to hashes tied to
the substructures at their respective feature-keys.

C<Data::FStructure::TiedHash> inherits from and supports all
methods provided by C<Data::FStructure>, including the automated
garbage-collection routines inherited from C<Data::FStructure::Safe>.

=cut


#--------------------------------------------------------------
# Usage
#--------------------------------------------------------------
=pod

=head1 USAGE

=cut


#--------------------------------------------------------------
# Tying
#--------------------------------------------------------------
=pod

=head2 Tying Up an FStructure

  my %args = { fs => $fs, hierarchy => $hi, ... };
  my %h = ();
  $tied = tie %h, Data::FStructure::TiedHash, \%args;

Ties the hash C<%h> to a new C<Data::FStructure::TiedHash> object.
Options supported by C<\%args> include all options supported by
C<Data::FStructure::new()>, plus the additional keyword-argument
C<fs>.

If specified, C<$fs> indicates any existing C<Data::FStructure>
object to which the hash C<%h> should be tied.  An B<exact>
copy (as created by the C<substructure()> method) of C<$fs> will be
made, promoted to the class
C<Data::FStructure::TiedHash>, and returned if this is the case.

If C<$fs> is not specified, an empty new Data::FStructure::TiedHash
object is created and returned.

All C<Data::FStructure::TiedHash> objects B<must> be rooted if
they are to function correctly -- a root node is automatically
created and assigned (if none is already present) by the
call to C<tie>.  The tied interface will cease to function
if for some reason the root node of the underlying
FStructure ceases to exist, (such as calling
C<remove_unreachable()> from a non-root node).

=cut


#--------------------------------------------------------------
# Assigning type-values
#--------------------------------------------------------------
=pod

=head2 Assigning Type-Values

Assigning a simple scalar value to any (nested) hash-key
assigns that scalar as a type to the substructure tied
to that hash-key's value:

  my $h = {};
  $fs = tie %$h, Data::FStructure::TiedHash;
  $h->{syn}{cat} = 'verb';
  $h->{syn} = 'vsyn';
  print $fs->as_pretty();

Prints something like:

  [ syn: [ (vsyn)          ] ]
  [      [ cat: [ (verb) ] ] ]

=cut

#'
#--------------------------------------------------------------
# Getting substructure values
#--------------------------------------------------------------
=pod

=head2 Getting Substructure Values

Continuing the example above, a reference to a hash tied to the
substructure at a given feature can be accessed via the built-in
perl hash-lookup syntax:

  use Data::FStructure::pretty;    # for pretty-printing

  $sub = $h->{syn};
  print tied(%$sub)->as_pretty();

Prints something like:

  [ (vsyn)          ]
  [ cat: [ (verb) ] ]

Note that non-existant paths are automagically created by such
lookups:

  $sub = $h->{sem};
  print tied(%$h)->as_pretty();

Prints something like:

  [ sem: [ ]                 ]
  [ syn: [ (vsyn)          ] ]
  [      [ cat: [ (verb) ] ] ]

=cut

#'
#--------------------------------------------------------------
# Setting substructure values
#--------------------------------------------------------------
=pod

=head2 Setting Substructure Values

Substructure values can be set by assigning any Data::FStructure
object to a tied-hash key:

  # make the substructures at 'sem' and 'syn' shared
  $h->{sem} = $h->{syn};
  print tied(%$h)->as_pretty;

Prints something like:

  [ sem: [ =1 (vsyn)          ] ]
  [      [ cat: [ =2 (verb) ] ] ]
  [ syn: [ =1 ]                 ]

As the above example shows, structure-sharing can be achieved
by simply assigning the hash-value of one path to that of
another path.  This is just one possibility for substructure-assignment.
Value-storage relies internally on the C<gobble()>
method provided by the C<Data::FStructure::perldata> module.
For details on what kinds of values can occur on the right-hand-side
of an assignment and how they are interpreted, see
L<Data::FStructure::perldata>.


=cut


#--------------------------------------------------------------
# Iterating
#--------------------------------------------------------------
=pod

=head2 Iterating

The built-in perl functions C<keys()>, C<values()>, and C<each()> work
as you would expect:

  while (($ftr,$hash) = each(%$h)) {
    print "F\@$ftr =\n", tied(%$hash)->as_pretty(2), "\n";
  }

Prints something like:

  F@sem =
    [ (vsyn)          ]
    [ cat: [ (verb) ] ]
  F@syn =
    [ (vsyn)          ]
    [ cat: [ (verb) ] ]

=cut


#--------------------------------------------------------------
# Accessing underlying FStructure objects
#--------------------------------------------------------------
=pod

=head2 Accessing the underlying FStructure object

Since Data::FStructure::TiedHash inherits from Data::FStructure,
the underlying FStructure object itself is returned by the
call to C<tie()>.  It can be recovered (in case you didn't
save it) by calling perl's built-in C<tied()> function on
any tied hash:

  # tie up a new FStructure
  $hphon = {};
  tie %$hphon, Data::FStructure::TiedHash;

  # give it some values
  $hphon->{phon}{pform} = '/mU/';

  # call an FStructure method (destructive unification)
  tied(%$hphon)->_unify(tied(%$h));

  # ... and see what happened
  print tied(%$hphon)->as_pretty();

Prints something like:

  [ phon: [ pform: [ (/mU/) ] ] ]
  [ sem: [ =1 (vsyn)          ] ]
  [      [ cat: [ =2 (verb) ] ] ]
  [ syn: [ =1 ]                 ]


=cut


#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Data::FStructure(3pm).
