# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::arydelta.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Feature-Structures for perl: delta function
#
###############################################################

package Data::FStructure::arydelta;
#use Carp;
use Exporter;
use Data::FStructure::delta;
use List::Util qw(max);
@ISA = qw(Data::FStructure::delta);
@EXPORT_OK = qw($ARRAY_FTR);
%EXPORT_TAGS = ( 'arrayvars' => [qw($ARRAY_FTR)] );

our $VERSION = 0.02;

our $ARRAY_FTR = 'FSARRAY';

###############################################################
# Constructor
#   + Data::FStructure::arydelta->new()
#   + Object-structure:
#     { qfrom1 => { ftr1a => qto1a, ... }
#       qarray1 => { $ARRAY_FTR => [ tonode0, tonode1, ... ] } }
#   + Args: NONE
###############################################################
# new : inherited

###############################################################
# warnings
###############################################################

# undef = warn_notnum($op) -- let perl do this for us
# --> actually, this is really ignored:
#     non-numeric features treated as "normal" ones
#     for arydelta v. > 0.01.
sub warn_notnum {
  carp("Warning: non-numeric feature in Data::FStructure::arydelta::$_[0]");
  return undef;
}

###############################################################
# Lookup
###############################################################

# \@ary = $dt->elts($from);
sub elts {
  return
    (defined($_[1]) && exists($_[0]->{$_[1]})
     ? (exists($_[2])
	? ($_[0]->{$_[1]}{$ARRAY_FTR} = $_[2])
	: $_[0]->{$_[1]}{$ARRAY_FTR})
     : undef);
}

# $to = $dt->fetch($from,$on);
sub fetch {
  return
    (defined($_[1]) && defined($_[2])
     ? (($_[2] =~ /^\d+$/)
	? (exists($_[0]->{$_[1]}{$ARRAY_FTR})
	   ? ($_[0]->{$_[1]}{$ARRAY_FTR}[$_[2]])
	   : undef)
	: $_[0]->{$_[1]}{$_[2]})
     : undef);
}

# @labels = $dt->labels_from($node)
sub labels_from {
  return
    ((grep { $_ ne $ARRAY_FTR } $_[0]->SUPER::labels_from($_[1])),
     (defined($_[1]) && exists ($_[0]->{$_[1]}) && exists($_[0]->{$_[1]}{$ARRAY_FTR})
      # brute force, but should be safe
      ? (grep { exists($_[0]->{$_[1]}{$ARRAY_FTR}[$_]) } 0..$#{$_[0]->{$_[1]}{$ARRAY_FTR}})
      : qw()));
}
# @alphabet = $dt->alphabet();
sub alphabet {
  return
    ((grep { $_ ne $ARRAY_FTR } $_[0]->SUPER::alphabet),
     0..max(map {
             exists($_->{$ARRAY_FTR})
	      ? $#{$_->{$ARRAY_FTR}}
	      : -1
	    } values(%{$_[0]})));
}


# $bool = $dt1->is_eq($dt2)
#sub is_eq { return defined($_[1]) && $_[0] eq $_[1]; } -- inherited


###############################################################
# Manipulation
###############################################################

# $dt = $dt->replace(%old2new)
sub replace {
  my ($dt,%rename) = @_;
  $dt->SUPER::replace(%rename);

  # rename array-node sources (array-transitions from) [done by super]

  # rename array-node elements (array-transitions to)
  my ($ary);
  foreach $ary (map {
                  exists($_->{$ARRAY_FTR})
		    ? $_->{$ARRAY_FTR}
		    : qw()
		} values(%$dt))
    {
      @$ary = map { exists($rename{$_}) ? $rename{$_} : $_ } @$ary;
    }

  return $dt;
}

# $to = $dt->store($from,$on,$to);
sub store {
  return
    (defined($_[1]) && defined($_[2])
     # interpret all numeric features as array-indices
     ? (($_[2] =~ /^\d+$/)
	? ($_[0]->{$_[1]}{$ARRAY_FTR}[$_[2]] = $_[3])
	# other features are treated normally
	: ($_[0]->{$_[1]}{$_[2]} = $_[3]))
     : undef);
}

# $dt = $dt->remove_transits_from(@nodes) -- inherited

# $dt = $dt->remove_transits_to(@nodes)
sub remove_transits_to {
  my $dt = shift;
  my %remove = map { $_ => undef } @_;

  my ($q,$ary);
  foreach $q (keys(%$dt)) {
    # could be inherited, but why build the hash twice?
    delete(@{$dt->{$q}}{grep { exists($remove{$dt->{$q}{$_}}) } keys(%{$dt->{$q}}) });
    if (defined($ary = $dt->{$q}{$ARRAY_FTR})) {
      delete(@$ary[grep { defined($ary->[$_]) && exists($remove{$ary->[$_]}) } 0..$#$ary]);
    }
  }
  return $dt;
}


# $dt = $dt->remove_transit($from,$on)
sub remove_transit {
  if (defined($_[2])
      # treat numeric features as array-indices
      && ($_[2] = ~ /^\d+$/)
      && defined($_[1])
      && exists($_[0]->{$_[1]}{$ARRAY_FTR})
      && exists($_[0]->{$_[1]}{$ARRAY_FTR}[$_[2]]))
    {
      delete($_[0]->{$_[1]}{$ARRAY_FTR}[$_[2]]);
      delete($_[0]->{$_[1]}{$ARRAY_FTR}) unless (@{$_[0]->{$_[1]}{$ARRAY_FTR}});
    }
  return $_[0]->SUPER::remove_transit($_[1],$_[2]);
}

# $dst_dt = $dst_dt->assign($src_dt);  -- inherited: nope
sub assign {
  #my ($dst,$src) = @_;
  %{$_[0]} = ();
  foreach (keys(%{$_[1]})) {
    %{$_[0]->{$_}} = %{$_[1]->{$_}};
    # handle array-ref pseudofeature -- make a new reference
    if (exists($_[1]->{$_}{$ARRAY_FTR})) {
      $_[0]->{$_}{$ARRAY_FTR} = [@{$_[1]->{$_}{$ARRAY_FTR}}];
    }
  }
  return $_[0];
}


# $dst_dt = $dst_dt->_merge($src_dt);
# ---> overrides existing transitions! (including pseudo-transitions)
# ---> should be safe for node-disjoint deltas
sub _merge {
  my ($dst,$src) = @_;
  foreach (keys(%$src)) {
    %{$dst->{$_}} = %{$src->{$_}};
    # handle array-ref pseudofeature -- make a new reference
    if (exists($src->{$_}{$ARRAY_FTR})) {
      $dst->{$_}{$ARRAY_FTR} = [@{$src->{$_}{$ARRAY_FTR}}];
    }
  }
  return $dst;
}


# $dst_dt = $dst_dt->merge($src_dt);
# ---> safe(r)
sub merge {
  my ($k,$i);
  foreach $k (keys(%{$_[1]})) {
    # old-style
    %{$_[0]->{$k}} = (exists($_[0]->{$k})
		      ? (%{$_[0]->{$k}}, %{$_[1]->{$k}})
		      : (%{$_[1]->{$k}}));
    # handle array-ref pseudofeature -- make a new reference
    if (exists($src->{$k}{$ARRAY_FTR})) {
      foreach $i (0..$#{$src->{$k}{$ARRAY_FTR}}) {
	$dst->{$k}{$ARRAY_FTR}[$i] = $src->{$k}{$ARRAY_FTR}[$i];
      }
    }
  }
  return $_[0];
}

# $dt = $dt->clear();
sub clear { %{$_[0]} = (); return $_[0]; }


1;
__END__

# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::arydelta -
Feature structures for perl -- delta functions for arrays.

=head1 SYNOPSIS

  use Data::FStructure::arydelta;

  #
  # Package Variables
  #
  $ARRAY_FTR = 'FSARRAY';                   # what to call our pseudo-feature

  $dt = Data::FStructure::arydelta->new();   # return a new delta function
  $dt->store($from,$i,$val);                # array-storage for numeric features
  $dt->fetch($from,$i);                     # ... retrieval for numeric features

  $elts = $dt->elts($from);                 # get perl array-reference of nodes

=head1 REQUIREMENTS

Exporter,
Data::FStructure::delta,
List::Util.

=cut

#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

C<Data::FStructure::arydelta> implements arrays for FStructure
objects as perl array-references at the value of a user-specifiable
pseudo-feature.  Numeric symbol- (feature-, label-, attribute-, ...)
names are automatically interpreted as array-indices.

=cut

#--------------------------------------------------------------
# Variables
#--------------------------------------------------------------
=pod

=head1 PACKAGE VARIABLES

The C<Data::FStructure::arydelta> package defines one
exportable variable,

  $ARRAY_FTR

which represents the name of the pseudo-feature under
which perl array-refernces are to be stored.  The
default value is C<'FSARRAY'>.  This pseudo-feature is
not reported by the usual C<labels_from()> or C<alphabet()>
functions, but can be accessed with C<fetch()> and
C<store()>, if desired (although it's better is to use the C<elts()>
method).

=cut

#'

#--------------------------------------------------------------
# Methods
#--------------------------------------------------------------
=pod

=head1 METHODS

C<Data::FStructure::arydelta> inherits from
C<Data::FStructure::delta>, and supports all methods
described in L<Data::FStructure::delta>, automatically
interpreting numeric feature-names as array-indices.

It I<is> possible (and safe) to have both numeric and symbolic
features shared by a single node, although this is
not recommended if you intend to use the C<spit()> method.

=over 4

=item * C<elts($from)>

Returns the perl array reference associated with the node
C<$from>, or C<undef> if C<$from> does not have an array
reference currently associated with it (it will be automagically
created if you assign a numeric feature-value to C<$from>).

=item * C<elts($from,\@ary)>

Sets the perl array reference associated with the node
C<$from> to C<\@ary>.

=back

=cut


#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Data::FStructure(3pm).
Data::FStructure::theta(3pm).
Data::FStructure::delta(3pm).
