# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::delta.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Feature-Structures for perl: delta function
#
###############################################################

package Data::FStructure::delta;

our $VERSION = 0.02;

###############################################################
# Constructor
#   + Data::FStructure::delta->new()
#   + Object-structure:
#     { qfrom1 => { ftr1a => qto1a, ... }, ... }
#   + Args: NONE
###############################################################
sub new { return bless {}, ref($_[0]) || $_[0]; }

###############################################################
# Lookup
###############################################################

# $to = $dt->fetch($from,$on);
sub fetch {
  return
    (defined($_[1]) && defined($_[2]) && exists($_[0]->{$_[1]})
     ? $_[0]->{$_[1]}{$_[2]}
     : undef);
}

# @labels = $dt->labels_from($node)
sub labels_from {
  return
    (defined($_[1]) && exists($_[0]->{$_[1]})
     ? (keys(%{$_[0]->{$_[1]}}))
     : qw());
}

# @alphabet = $dt->alphabet();
sub alphabet { return (map { keys(%$_) } values(%{$_[0]})); }


# $bool = $dt1->is_eq($dt2)
sub is_eq { return defined($_[1]) && $_[0] eq $_[1]; }


###############################################################
# Manipulation
###############################################################

# $dt = $dt->replace(%old2new)
sub replace {
  my ($dt,%rename) = @_;

  # delete transits-from
  @$dt{values(%rename)} = @$dt{keys(%rename)};
  delete(@{$dt}{keys(%rename)});

  # delete transits-to
  my ($q);
  foreach $q (keys(%$dt)) {
    @{$dt->{$q}}{keys(%{$dt->{$q}})} =
      map {
	exists($rename{$_}) ? $rename{$_} : $_
      } values(%{$dt->{$q}});
  }
  return $dt;
}



# $to = $dt->store($from,$on,$to);
sub store {
  return
    (defined($_[1]) && defined($_[2])
     ? $_[0]->{$_[1]}{$_[2]} = $_[3]
     : undef);
}

# $dt = $dt->remove_transits_from(@nodes)
sub remove_transits_from {
  delete(@{$_[0]}{@_[1..$#_]});
  return $_[0];
}

# $dt = $dt->remove_transits_to(@nodes)
sub remove_transits_to {
  my $dt = shift;
  my %remove = map { $_ => undef } @_;
  my ($q);
  foreach $q (keys(%$dt)) {
    delete(@{$dt->{$q}}{grep { exists($remove{$dt->{$q}{$_}}) } keys(%{$dt->{$q}}) });
  }
  return $dt;
}

# $dt = $dt->remove_transit($from,$on)
sub remove_transit {
  if (defined($_[1]) && defined($_[2]) && exists($_[0]->{$_[1]})) {
    delete($_[0]->{$_[1]}{$_[2]});
    delete($_[0]->{$_[1]}) unless (%{$_[0]->{$_[1]}});
  }
  return $_[0];
}

# $dst_dt = $dst_dt->assign($src_dt);
sub assign {
  %{$_[0]} = ();
  foreach (keys(%{$_[1]})) {
    %{$_[0]->{$_}} = %{$_[1]->{$_}};
  }
  return $_[0];
}

# $dst_dt = $dst_dt->_merge($src_dt);
# ---> overrides existing transitions!
sub _merge {
  foreach (keys(%{$_[1]})) {
    %{$_[0]->{$_}} = %{$_[1]->{$_}};
  }
  return $_[0];
}

# $dst_dt = $dst_dt->merge($src_dt);
# ---> safe
sub merge {
  foreach (keys(%{$_[1]})) {
    %{$_[0]->{$_}} =
      (exists($_[0]->{$_})
       ? (%{$_[0]->{$_}}, %{$_[1]->{$_}})
       : (%{$_[1]->{$_}}));
  }
  return $_[0];
}

# $dt = $dt->clear();
sub clear { %{$_[0]} = (); return $_[0]; }

1;
__END__



# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::delta -
Feature structures for perl -- transition function "delta".

=head1 SYNOPSIS

  use Data::FStructure::delta;

  #
  # Methods
  #
  $dt = Data::FStructure::delta->new();     # new transition-map

  $to = $dt->fetch($from,$on);             # get transition-value
  $to = $dt->store($from,$on,$to);         # set transition-value

  @labels = $dt->labels_from($node);       # get available transition-labels
  @alphabet = $dt->alphabet();             # get all defined transition-labels

  $dt = $dt->replace($old=>$new,...);      # rename nodes
  $dt = $dt->remove_transits_from(@nodes); # remove any transitions from @nodes
  $dt = $dt->remove_transit($from,$on);    # remove single transition

  $bool = $dt1->is_eq($dt2);               # test object-identity
  $dt_dst = $dt_dst->assign($dt_src);      # assign $src_dt to $dst_dt
  $dt_dst = $dt_dst->merge($dt_src);       # merge/override (safe)
  $dt_dst = $dt_dst->_merge($dt_src);      # merge/override (less safe)
  $dt = $dt->clear();                      # drop all transitions


=head1 REQUIREMENTS

Just perl.

=cut

#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

C<Data::FStructure::delta> objects provide type-mappings to
FStructure objects.  Currently,
C<Data::FStructure::delta> objects are implemented as 2-level perl hashes.

=cut

#--------------------------------------------------------------
# Variables
#--------------------------------------------------------------
#=pod
#
#=head1 PACKAGE VARIABLES
#
#=cut

#--------------------------------------------------------------
# Methods
#--------------------------------------------------------------
=pod

=head1 METHODS

=over 4

=item * C<new()>

Returns a new C<Data::FStructure::delta> object.

=item * C<fetch($from,$on)>

Returns the transition-value from the node C<$from>
at the symbol (read: "feature") C<$on>, or C<undef>
if not defined.

=item * C<store($from,$on,$to)>

Stores the transition-value from the node C<$from>
at the symbol C<$on> to the node C<$to>.
C<$from> and C<$on> must be defined.  Returns C<$to>.

=item * C<labels_from($node)>

Returns the labels (symbols, features, ...) currently
defined as transitions from the node C<$node> as a
list.

=item * C<alphabet()>

Returns all labels (symbols, features, ...) currently
defined as transitions from any node as a list.

=item * C<< replace($old=E<gt>$new,...) >>

Renames the C<$old> nodes to their C<$new> mappings.
Both transitions from- and to- the C<$old> nodes are
renamed.  Returns the C<delta> object.

=item * C<remove_transits_from(@nodes)>

Removes all transitions currently defined for
any of the nodes in C<@nodes>.  Returns the C<delta>
object.

=item * C<remove_transit($from,$on)>

Removes the transition from the node C<$from> on
the feature C<$on>, if defined.  Returns the
C<delta> object.

=item * C<< $dt1-E<gt>is_eq($dt2) >>

Returns a true value if C<$dt1> is object-identical
to C<$dt2>, false otherwise.  Currently, this just checks
string-identity with C<eq>, but that should be pretty safe.

=item * C<< $dt_dst-E<gt>assign($dt_src) >>

Makes C<$dt_dst> an information-identical (but object-distinct)
copy of C<$dt_src>, returns C<$dt_dst>.

=item * C<< $dt_dst-E<gt>merge($dt_src) >>

Merges transitions from C<$dt_src> into C<$dt_dst>.
Existing transitions in C<$ds_dst> may get clobbered
if C<$dt_src> and C<$dt_dst> share node- and symbol-names.
Returns C<$dt_dst>.

=item * C<< $dt_dst-E<gt>_merge($dt_src) >>

Like C<merge()>, but existing transitions in C<$dt_dst>
may not only be clobbered, but may in fact be deleted if
any node- and symbol-names are shared.
If no node-names are shared, this method should be safe.
Returns C<$dt_dst>.

=item * C<clear()>

Clears all transitions.

=back

=cut


#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

Bob Carpenter, I<The Logic of Typed Feature Structures>.
Cambridge University Press, 1992.

perl(1).
Data::FStructure(3pm).
Data::FStructure::theta(3pm).
Data::FStructure::arydelta(3pm).
