# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::pretty.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Feature-structures for perl : pretty viewing
#
###############################################################

package Data::FStructure::pretty;
use Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = (
	      qw($PATH_SEPARATOR $PATH_SPLIT_REGEX),
	      qw($PS_VIEWER $DVI_VIEWER $LATEX_PROGRAM),
#	      qw(@TMPFILES $UNLINK_TMPFILES)
	     );
%EXPORT_TAGS =
  (
   'pathvars' => [ qw($PATH_SEPARATOR $PATH_SPLIT_REGEX) ],
   'progvars' => [ qw($PS_VIEWER $DVI_VIEWER $LATEX_PROGRAM) ],
#   'tmpvars' => [ qw(@TMPFILES $UNLINK_TMPFILES) ]
  );


###############################################################
# Package-global variables
###############################################################
our $VERSION = 0.02;

our $PATH_SEPARATOR = ':';
our $PATH_SPLIT_REGEX = ':';

our $PS_VIEWER = $Math::PartialOrder::Loader::PS_VIEWER || 'ghostview';
our $DVI_VIEWER = 'xdvi';
our $LATEX_PROGRAM = 'latex';

our @TMPFILES = qw();
our $UNLINK_TMPFILES = 1;

###############################################################
# Initialization:
# $fs->initialize(\%args)
# + new keys:
#     pathsep => $str,
#     pathsplit => $regex,
###############################################################
sub initialize {
  my ($fs,$args) = @_;
  $fs->{pathsep} = $args->{pathsep} if exists($args->{pathsep});
  $fs->{pathsplit} = $args->{pathsplit} if exists($args->{pathsplit});
  return $fs;
}

###############################################################
#
###############################################################



###############################################################
# sneakily switch packages
###############################################################
package Data::FStructure::Base;
use Data::Dumper;
use GraphViz;
use Text::Balanced qw(extract_bracketed);
use File::Temp qw(tempfile);
Data::FStructure::pretty->import qw(:pathvars :progvars);


###############################################################
# path/string utilities
###############################################################

#--------------------------------------------------------------
# accessors/manipulators
#--------------------------------------------------------------

# pathsep(), pathsep($string)
# Alias: path_separator
*path_separator = \&pathsep;
sub pathsep {
  return
    (exists($_[1])
     ? ($_[0]->{pathsep} = $_[1])
     : (defined($_[0]->{pathsep})
	? $_[0]->{pathsep}
	: $PATH_SEPARATOR));
}

# pathsplit(), pathsplit($string)
# Alias: path_split_regex
*pathsplit_regex = \&pathsplit;
*path_split_regex = \&pathsplit;
sub pathsplit {
  return
    (exists($_[1])
     ? ($_[0]->{pathsplist} = $_[1])
     : (defined($_[0]->{pathsplist})
	? $_[0]->{pathsep}
	: $PATH_SPLIT_REGEX));
}

#--------------------------------------------------------------
# high-level
#--------------------------------------------------------------

# @strings = $fs->path2string(@paths)
# ALIASES: paths2strings
sub paths2strings {
  my $fs = shift;
  my $sep = $fs->path_separator;
  return map { join($sep, @$_) } @_;
}
*path2string = \&paths2strings;


# @paths = $fs->string2path(@strings)
# ALIASES: strings2paths
sub strings2paths {
  my $fs = shift;
  my $re = $fs->path_split_regex;
  return map { [split($re, $_)] } @_;
}
*string2path = \&strings2paths;


# @pstrings = $fs->hookstrings();
sub hookstrings { return $_[0]->path2string($_[0]->hooks); }

# @pstrings = $fs->pathstrings();
sub pathstrings { return $_[0]->path2string($_[0]->paths); }




###############################################################
# Text Methods
###############################################################

#--------------------------------------------------------------
# text2fs
#--------------------------------------------------------------

# $fs = $fs->from_text($string);
# $fs = $fs->from_string($string)
*from_string = \&from_text;
sub from_text {
  my ($fs,$string) = @_;
  $fs->clear;
  $string =~ s/\n/ /g; # no embedded newlines!
  my %tags = ();                       # hash: node by tag-string
  my ($node,$rem,$tag,$type);

  # --- INITIALIZATION (root node) ---
  ($string,$rem) = extract_bracketed($string, '[]');
  return $fs if ($string =~ /^\s*$/); # empty string == unrooted fs

  # new root node
  $node = $fs->_gensym_node();
  $fs->add_node($node);
  $fs->root($node);

  # remove open-bracket
  $string =~ s/^\[\s*//;

  # get tag
  $tag = $string =~ s/^=(\w+)\s*// ? $1 : undef;
  $tags{$tag} = $node if (defined($tag));

  # get type
  ($type,$string) = extract_bracketed("$string", '()');
  if ($type) {
    $type =~ s/^\(//;
    $type =~ s/\)$//;
    $fs->theta($node,$type) if ($type);
  }

  my @nodeq = ($node);                 # queue of nodes
  my @stringq = ($string);             # parallel queue of strings
  my $re = $fs->path_split_regex;     # path-separator regexp
  my (@extracted,$ftr,$val,$newnode);
  while (defined($node = shift(@nodeq))) {
    $string = shift(@stringq);
    next unless ($string); # empty node

    # process features
    while (@extracted = extract_bracketed
	   ("$string",
	    '[]',
	    "\\s*\\S+\\s*${re}\\s*")) {
      ($val,$rem,$ftr) = @extracted;

      # propagate errors caught by 'extract_bracketed'
      if (!defined($val)) {
	carp("error parsing text feature-structure: $@");
	return undef;
      }
      last unless ($val);

      # extract feature-name
      $ftr =~ s/\s*${re}\s*$//;
      $ftr =~ s/^\s*//g;

      # remove open-bracket
      $val =~ s/^\[\s*//;

      # get tag
      $tag = $val =~ s/^=(\w+)\s*// ? $1 : undef;

      # get type
      ($type,$val) = extract_bracketed("$val", '()');
      if ($type) {
	$type =~ s/^\(//;
	$type =~ s/\)$//;
      }

      # do we need a new node?
      if (defined($tag)) {
	if (exists($tags{$1})) {
	  # nope.
	  $newnode = $tags{$1};
	} else {
	  # yup.
	  $newnode = $fs->_gensym_node();
	  $tags{$tag} = $newnode;
	}
      } else {
	# also yup.
	$newnode = $fs->_gensym_node();
      }

      # add node, theta-value, and transition-to-node
      $fs->add_node($newnode);
      $fs->theta($newnode,$type) if ($type ne '');
      $fs->delta($node,$ftr,$newnode);

      # add node & structure to queue
      push(@nodeq, $newnode);
      push(@stringq, $val);

      # keep going
      $string = $rem;
    }
  }
  return $fs;
}

#--------------------------------------------------------------
# fs2text
#--------------------------------------------------------------
# $string = $fs->dump()
sub dump { return Data::Dumper->Dump([$_[0]], [$_[0]]); }

# @strings = $fs->as_pathval_strings;
# (NYD)
sub as_pathval_strings {
  my $fs = shift;
  return
    @{$fs->iterate_hooks({
			  callback => \&_pathval_strings_callback,
			  pstrings => {},
			  return => [],
			 })};
}
sub _pathval_strings_callback {
  my ($fs, $path, $node, $args) = @_;
  my ($valstring);
  if (exists($args->{pstrings}->{$node})) {
    $valstring =
      join($fs->path_separator, @$path) . '=${' . $args->{pstrings}->{$node} . '}';
  } else {
    $valstring = $args->{pstrings}->{$node} = join($fs->path_separator, @$path);
    $valstring .=
      '=' . (defined($fs->theta($node))
	     ? $fs->theta($node)
	     : 'undef');
  }
  push(@{$args->{return}}, $valstring);
  return undef;
}

# $string = $fs->as_string()
sub as_string {
  my $fs = shift;
  return
    $fs->_as_tagged
      ({
	indentinc => {},
	pindent => {},
	cindent => 0,
	path_separator => $fs->path_separator,
	place_feature => \&_as_string_feature,
	place_openp => \&_as_text_openp,
	place_tag => \&_as_text_tag,
	place_type => \&_as_text_type,
	place_closep => \&_as_text_closep,
	return => ''
       })
}
# place_feature($ftr,$path,$node,$args)
sub _as_string_feature {
  my $args = $_[3];
  $args->{return} .= ' ' . $_[0] . $args->{path_separator} . ' ';
  $args->{indentinc}{$_[1]} = 0;
  $args->{pindent}{$_[1]} = 0;
}

# $text = $fs->as_text()
sub as_text {
  my $fs = shift;
  return
    $fs->_as_tagged
      ({
	indentinc => {},
	pindent => {},
	cindent => 0,
	path_separator => $fs->path_separator,
	place_feature => \&_as_text_feature,
	place_openp => \&_as_text_openp,
	place_tag => \&_as_text_tag,
	place_type => \&_as_text_type,
	place_closep => \&_as_text_closep,
	return => ''
       })
}
# place_feature($ftr,$path,$node,$args)
sub _as_text_feature {
  my $args = $_[3];
  $args->{return} .= "\n" . ' ' x $args->{cindent} . $_[0] . $args->{path_separator} . ' ';
  $args->{indentinc}{$_[1]} = length($_[0]) + length($args->{path_separator}) + 1;
  $args->{pindent}{$_[1]} = $args->{cindent} + $args->{indentinc}{$_[1]};
  $args->{cindent} = $args->{pindent}{$_[1]};
}
# place_openp($path,$node,$args)
sub _as_text_openp {
  my $args = $_[2];
  $args->{return} .=  '[';
  $args->{pindent}{$_[0]} += 2;
  $args->{indentinc}{$_[0]} += 2;
  $args->{cindent} += 2;
}
# place_tag($tag,$path,$node,$args)
sub _as_text_tag { $_[3]->{return} .= " =$_[0]" if (defined($_[0])); }
# place_type($type,$path,$node,$args)
sub _as_text_type { $_[3]->{return} .= " ($_[0])" if (defined($_[0])); }
# place_closep($path,$node,$args)
sub _as_text_closep {
  my $args = $_[2];
  $args->{return} .= ' ]';
  $args->{cindent} -= $args->{indentinc}{$_[0]} if (defined($args->{indentinc}{$_[0]}));
}

# $pretty_string = $fs->as_pretty($indent)
sub as_pretty {
  my $fs = shift;
  my $indent = shift || 0;
  $indent = ' ' x $indent;
  return $indent . join("\n$indent", $fs->as_pretty_list);
}
sub as_pretty_list {
  my $fs = shift;
  return qw() unless (defined($fs->root));
  return
    @{$fs->_as_tagged
	({
	  openposns => {},
	  closeposns => {},
	  openlines => {},
	  neednewline => {},
	  place_feature => \&_as_pretty_feature,
	  place_openp => \&_as_pretty_openp,
	  place_tag => \&_as_pretty_tag,
	  place_type => \&_as_pretty_type,
	  place_closep => \&_as_pretty_closep,
	  return => [''],
	 })};
}
# place_feature($ftr,$path,$node,$args)
sub _as_pretty_feature {
  my ($ftr,$path,$args) = ($_[0],$_[1],$_[3]);
  # get prefix-open-position (+2 spaces for openp & pad-space)
  my $indent = $args->{openposns}{$args->{prefixof}{$path}} + 2;
  # set up newline if necessary
  if ($args->{neednewline}{$args->{prefixof}{$path}}) {
    push(@{$args->{return}}, ' ' x $indent);
  } else {
    $args->{return}[$#{$args->{return}}] .= ' ';
  }
  $args->{neednewline}{$args->{prefixof}{$path}} = 1;
  # adjust and set open-position
  $args->{openposns}{$path} = $indent + length($ftr) + 2;
  # add feature-string
  $args->{return}[$#{$args->{return}}] .= "$ftr: ";
}
# place_openp($path,$node,$args)
sub _as_pretty_openp {
  my $args = $_[2];
  my $lines = $args->{return};
  # remember opening line number & offset
  $args->{openlines}{$_[0]} = $#$lines;
  $args->{openposns}{$_[0]} = 0 unless ($args->{openposns}{$_[0]});
  # place the open-bracket
  $lines->[$#$lines] .= '[';
}
# place_tag($tag,$path,$node,$args)
sub _as_pretty_tag {
  if (defined($_[0])) {
    $_[3]->{return}[$#{$_[3]->{return}}] .= " =$_[0]";
    $_[3]->{neednewline}{$_[1]} = 1;
  }
}
# place_type($type,$path,$node,$args)
sub _as_pretty_type {
  if (defined($_[0])) {
    $_[3]->{return}[$#{$_[3]->{return}}] .= " ($_[0])";
    $_[3]->{neednewline}{$_[1]} = 1;
  }
}
# place_closep($path,$node,$args)
sub _as_pretty_closep {
  my ($path,$args) = ($_[0],$_[2]);
  my $lines = $args->{return};
  my $openpos = $args->{openposns}{$path};
  my ($i);
  # get close-position
  my $closepos = $args->{closeposns}{$path} || 0;
  for ($i = $args->{openlines}{$path}; $i <= $#$lines; ++$i) {
    $closepos = length($lines->[$i]) + 1 unless
      ($closepos >= length($lines->[$i])+1);
  }
  # extend brackets
  for ($i = $args->{openlines}{$path}; $i <= $#$lines; ++$i) {
    $lines->[$i] .= ' ' x ($closepos - length($lines->[$i]));
    substr($lines->[$i], $openpos, 1) = '[';
    substr($lines->[$i], $closepos, 1) = ']';
  }
  # save close-position
  $args->{closeposns}{$path} = $closepos;
  # adjust close-position of prefix
  my $prefix = $args->{prefixof}{$path};
  if ($prefix) {
    $args->{closeposns}{$prefix} = $closepos + 2
      unless (defined($args->{closeposns}{$prefix}) &&
	      $args->{closeposns}{$prefix} >= ($closepos + 2));
  }
}


###############################################################
# LaTeX
###############################################################

# $avmstring = $fs->as_latex_avm
sub as_latex_avm {
  my $fs = shift;
  return
    "\\begin{avm}" .
    $fs->_as_tagged
      ({
	indentinc => 3,
	indent => 3,
	neednewline => {},
	place_feature => \&_latex_avm_feature,
	place_openp => \&_latex_avm_openp,
	place_tag => \&_latex_avm_tag,
	place_type => \&_latex_avm_type,
	place_closep => \&_latex_avm_closep,
       })
    . "\n\\end{avm}\n";
}
# place_feature($ftr,$path,$node,$args)
sub _latex_avm_feature {
  my ($path,$args) = ($_[1],$_[3]);
  # set up newline if necessary
  $args->{return} .= "\\cr" if
    ($args->{neednewline}{$args->{prefixof}{$path}});

  $args->{neednewline}{$args->{prefixof}{$path}} = 1;

  $args->{return} .=  "\n" . ' ' x $_[3]->{indent}
    . "\\ensuremath\{\\textsc\{$_[0]\}\}: &";
}
# place_openp($path,$node,$args)
sub _latex_avm_openp {
  $_[2]->{return} .= "\n" . ' ' x $_[2]->{indent} . '\[';
  $_[2]->{indent} += $_[2]->{indentinc};
}
# place_tag($tag,$path,$node,$args)
sub _latex_avm_tag {
  if (defined($_[0])) {
    $_[3]->{neednewline}{$_[1]} = 1;
    $_[3]->{return} .= " \\\@\{$_[0]\}";
  }
}
# place_type($type,$path,$node,$args)
sub _latex_avm_type {
  if (defined($_[0])) {
    $_[3]->{neednewline}{$_[1]} = 1;
    $_[3]->{return} .= " \\ensuremath\{\\textbf\{$_[0]\}\}";
  }
}
# place_closep($path,$node,$args)
sub _latex_avm_closep {
  $_[2]->{return} .= ' \]';
  $_[2]->{indent} -= $_[2]->{indentinc};
}



###############################################################
# Graphviz
###############################################################

# $graphviz = as_graphviz()
sub as_graphviz {
  my ($fs,%opts) = @_;
  my $g = GraphViz->new(directed => 1,
			rankdir => 1, # left->right linking
			#node => {},
			#edge => {},
			%opts);
  my ($node,$type,$ftr);
  # add nodes
  foreach $node ($fs->nodes) {
    $type = defined($fs->theta($node)) ? $fs->theta($node) : 'undef';
    $g->add_node($node,
		 label => "$node: $type"
		 #label => "$type"
		);
  }
  # add edges
  foreach $node ($fs->nodes) {
    foreach $ftr ($fs->features($node)) {
      $g->add_edge($node,
		   $fs->delta($node,$ftr),
		   label => "$ftr");
    }
  }
  # add root-node notation
  if (defined($fs->root)) {
    my $empty = $fs->_gensym_node;
    $g->add_node($empty, shape => 'plaintext', label => '');
    $g->add_edge($empty, $fs->root);
  }
  return $g;
}

###############################################################
# Handy shortcuts
###############################################################
*gv = \&viewps;
sub viewps {
  my $fs = shift;
  my ($fh,$filename) = tempfile('fsXXXXXX', SUFFIX => '.ps');
  $fh->print($fs->as_graphviz->as_ps);
  close($fh);
  system("$PS_VIEWER \"$filename\" &");
  push(@Data::FStructure::pretty::TMPFILES,$filename) if ($Data::FStructure::pretty::UNLINK_TMPFILES);
}

# $fs->xdvi
*xdvi = \&viewdvi;
sub viewdvi {
  my $fs = shift;
  my ($fh,$filename) = tempfile('fsXXXXXX', SUFFIX => '.tex');
  $fh->print
    ('\documentclass[12pt]{article}', "\n",
     '\usepackage{avm}', "\n",
     '\begin{document}', "\n",
     $fs->as_latex_avm,
     '\end{document}', "\n");
  $fh->close;
  system("$LATEX_PROGRAM \"$filename\"");
  # cleanup those ugly .log and .aux files right away...
  substr($filename, -4,4,'');
  unlink("$filename.aux", "$filename.log");
  system("xdvi $filename.dvi &");
  if ($Data::FStructure::pretty::UNLINK_TMPFILES) {
    push(@Data::FStructure::pretty::TMPFILES, "$filename.tex", "$filename.dvi");
  }
}


###############################################################
# Low-level Utilities
###############################################################

# $fs->_as_tagged(\%args)
#  + abstract method for building tagged structure representation
#    for $fs; construction specified by \%args
#  + maintains the following keys of \%args
#    - open_nodes => $open
#       hashref of open nodes:
#       $open->{$node} > 0 iff $node has been opened (see 'place_openp')
#    - closed_nodes => $closed
#       hashref of closed nodes
#       $closed->{$node} > 0 iff $node has been closed (see 'place_closep')
#    - ignoreprefixes => Set::Hashed
#       set of prefixes of paths whose nodes have already been
#        closed over some (other) path
#    - tags => $tags
#       $tags->{$node} defined & unique iff $node is shared in $fs
#  + observes the following keys of \%args
#    - place_feature => $sub
#       called to add a feature to the path currently being constructed
#       calling convention:
#        $sub($ftr,$path,$node,\%args)
#    - place_tag => $sub
#       called to add a tag to the substructure currently being constructed
#       calling convention:
#        $sub($tag,$path,$node,\%args)
#    - place_type => $sub
#       called to add a type-label to the substructure currently being constructed
#       calling convention:
#        $sub($type,$path,$node,\%args)
#    - place_openp => $sub
#       called to start constructing a new substructure
#       calling convention:
#        $sub($path,$node,\%args)
#    - place_closep => $sub
#       called to stop constructing the current substructure
#       calling convention:
#        $sub($path,$node,\%args)
#  + none of the 'place_*' subroutines will be called
#    for paths leading away from closed nodes (nodes for
#    which 'place_closep' has been called).
#  + 'place_type' will only be called for the first hook
#    found to reach a given node: once a node is closed
#    (with the 'place_closep' subroutine), via some path,
#    all other paths reaching it are ignored.
#  + all keys maintained by iterate_tracking_prefixes() are
#   also available to callbacks (underlying implementation)
sub _as_tagged {
  my $fs = shift;
  my $args = shift;
  my $tags = {};
  $fs->iterate_hooks
    ({
      callback => \&_gather_tags_callback,
      firstpath => {},
      thetag => 0,
      tags => $tags,
     });
  return
    $fs->iterate_tracking_prefixes
      ({
	%$args,
	callback_open => \&_as_tagged_callback_open,
	callback_close => \&_as_tagged_callback_close,
	ignoreprefixes => Set::Hashed->new(),
	open_nodes => {},
	closed_nodes => {},
	tags => $tags
       });
}
sub _as_tagged_callback_open {
  my ($fs, $path, $node, $args) = @_;

  return undef if
    # --> ignoring some prefix of this path: skip it
    (!$args->{ignoreprefixes}->is_null &&
     !$args->{prefixes}{$path}->is_null &&
     (!$args->{prefixes}{$path}->
      intersection($args->{ignoreprefixes})->is_null));

  if (@$path && exists($args->{place_feature})) {
    # non-empty first-time path: place feature
    $args->{place_feature}($path->[$#$path],$path,$node,$args);
  }

  # open-substructure bracket
 if (exists($args->{place_openp})) {
   $args->{place_openp}($path,$node,$args);
  }

  # get tag if we need one
  if (exists($args->{place_tag})) {
    $args->{place_tag}($args->{tags}{$node},$path,$node,$args);
  }

  # multiply-occurring node from redundant access-path
  # --> ignore it and any suffixes!
  if ($args->{open_nodes}{$node} || $args->{closed_nodes}{$node}) {
    $args->{ignoreprefixes}->insert($path);
    return undef;
  }
  $args->{open_nodes}{$node} = 1;

  # we've never been here before: place the type
  if (exists($args->{place_type})) {
    $args->{place_type}($fs->theta($node),$path,$node,$args);
  }
  return undef;
}
sub _as_tagged_callback_close {
  my ($path, $node, $args) = @_[1..3];
  # remember closed nodes
  $args->{closed_nodes}{$node} = 0 unless ($args->{closed_nodes}{$node});
  ++$args->{closed_nodes}{$node};
  return undef if
    # --> ignoring some prefix of this path: skip it
    (!$args->{ignoreprefixes}->is_null &&
     !$args->{prefixes}{$path}->is_null &&
     (!$args->{prefixes}{$path}->
      intersection($args->{ignoreprefixes})->is_null));
  # close-bracket
  if (exists($args->{place_closep})) {
    $args->{place_closep}($path,$node,$args);
  }
  return undef;
}
sub _gather_tags_callback {
  my ($fs, $path, $node, $args) = @_;
  my $pstring = join($fs->path_separator, @$path);
  return undef if (exists($args->{tags}->{$node})); # already tagged
  if (exists($args->{firstpath}->{$node})) {
    $args->{tags}->{$node} = ++$args->{thetag};
  } else {
    $args->{firstpath}->{$node} = $pstring;
  }
  return undef;
}

END {
  #print STDERR "(cleanup)\n"; #join(', ', @Data::FStructure::pretty::TMPFILES), "\n";
  unlink(@Data::FStructure::pretty::TMPFILES)
    if ($Data::FStructure::pretty::UNLINK_TMPFILES);
}

1;
__END__

# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::pretty - Feature structures for perl -- pretty-printing.

=head1 SYNOPSIS

  use Data::FStructure::pretty;

  #
  # Package Variables
  #
  $PATH_SEPARATOR = ':';                         # for path-to-string conversion
  $PATH_SPLIT_REGEX = ':';                       # for string-to-path conversion

  $PS_VIEWER = 'ghostview';                      # for viewps() method
  $DVI_VIEWER = 'xdvi';                          # for viewdvi() method
  $LATEX_PROGRAM = 'latex';                      # for viewdvi() method

  @Data::FStructure::pretty::TMPFILES = qw();     # not exportable
  $Data::FStructure::pretty::UNLINK_TMPFILES = 1; # not exportable

  #
  # Construction & Initialization
  #
  $fs = Data::FStructure::Base->new(\%args);       # extra arguments

  #
  # Path/String Methods
  #
  $string = $fs->pathsep();              # get path-separator string
  $fs->pathsep($string);                 # set path-separator string

  $regex = $fs->pathsplit();             # get path-split regular expression
  $fs->pathsplit($regex);                # set path-split regular expression

  @strings = $fs->paths2stings(@paths);  # convert paths to strings
  @paths = $fs->strings2paths(@strings); # convert strings to paths

  @hookstrings = $fs->hookstrings();     # get all hooks() as strings
  @pathstrings = $fs->pathstrings();     # get all paths() as strings

  #
  # Text Methods
  #
  $fs->from_text($text);                 # read from text-string
  $text = $fs->as_text();                # output as text-string
  $string = $fs->as_string();            # ... with less whitespace

  $string = $fs->dump();                 # dump with Data::Dumper

  #
  # Other Viewing Methods
  #
  $avm_string = $fs->as_latex_avm();     # LaTeX code using the "avm" macros
  $g = $fs->as_graphviz();               # $g is a GraphViz object

  $fs->viewps();                         # PostScript viewing utility
  $fs->viewdvi();                        # DVI viewing utility

  #
  # Low-level Utilities
  #
  $fs->as_tagged(\%args);                # Generate a "tagged" structure


=head1 REQUIREMENTS

Carp,
Data::FStructure::Base,
Data::Dumper (for text dump),
Text::Balanced (for reading from strings),
File::Temp (for runtime viewing).

=cut

#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

C<Data::FStructure::pretty> defines methods in
the C<Data::FStructure::Base> package for FStructure viewing.

=cut

#--------------------------------------------------------------
# Variables
#--------------------------------------------------------------
=pod

=head1 PACKAGE VARIABLES

=over 4

=item * C<$PATH_SEPARATOR>

I<(Exportable)>

This string is used as the default value for
the path-separator string in path-to-string conversions.
The default value is ':'.

=item * C<$PATH_SPLIT_REGEX>

I<(Exportable)>

This regular expression is used as the default value for the
path-splitting regular expression in string-to-path conversions.
The default value is ':'.

=item * C<$PS_VIEWER>

I<(Exportable)>

Which PostScript viewer is to be used for the C<viewps()> method.
The default value is 'gv'.

=item * C<$DVI_VIEWER>

I<(Exportable)>

Which dvi viewer is to be used for the C<viewdvi()> method.
The default value is 'xdvi'.

=item * C<$LATEX_PROGRAM>

I<(Exportable)>

Which program is to be used to format the LaTeX code during
C<viewdvi()>.  The default value is 'latex'.

=item * C<$UNLINK_TMPFILES>

Whether to unlink temporary files generated by C<viewps()>
or C<viewdvi()> on exit.  The default value is C<1> (yes).

=item * C<@TMPFILES>

I<(Non-exportable)>

A list of temporary filenames generated by C<viewps()> or
C<viewdvi()>.  They will be removed on program exit
if C<$UNLINK_TMPFILES> is set to a true value.

=back

=cut

#--------------------------------------------------------------
# Methods
#--------------------------------------------------------------
=pod

=head1 METHODS

=cut


#--------------------------------------------------------------
# Construction & Initialization
#--------------------------------------------------------------
=pod

=head2 Construction and Initialization

=over 4

=item * C<initialize(\%args)>

Called by the (inherited) C<Data::FStructure::Base> method C<new()>;
additionally recognizes the following keys of C<\%args>:

=over 4

=item * C<< path_separator =E<gt> $string >>

Set the path-separator string for the new FStructure.

=item * C<< path_separator_re =E<gt> $regexp >>

Set the path-splitting regular expression for the new FStructure.

=back

=back

=cut


#--------------------------------------------------------------
# Path/String Methods
#--------------------------------------------------------------
=pod

=head2 Path/String Methods

=over 4

=item * C<paths2strings(@paths)>

Returns a list of string-representations of the feature-paths
in @paths.  Really just a glorified C<join()>.

See also: C<path_separator()>.

=item * C<strings2paths(@strings)>

Returns a list of feature-paths based on the feature-path string
representations @strings.  Really just a glorified C<split()>.

See also: C<path_separator_re()>.

=item * C<path_separator()>, C<path_separator($string)>

Get/set the string used to translate from paths to their
string representations.  For predictable results, this string
should never occur within an actual path.
The default value is C<$Data::FStructure::PATH_SEPARATOR>.

=item * C<path_separator_re()>, C<path_separator_re($string)>

Get/set the regular expression used to split strings representing
paths into their feature components.  For predictable results,
this expression should never match any other part of an
actual path-string.
The default value is C<$Data::FStructure::PATH_SEPARATOR_RE>.

=back

=cut


#--------------------------------------------------------------
# Text Methods
#--------------------------------------------------------------
=pod

=head2 Text Methods

=over 4

=item * C<from_text($string)>, C<from_string($string)>

Populates the FStructure based on the text-representation in
$string, which should be constructed according to the following
grammar:

  FSTRING  ::= '[' TAG? TYPE? FEATURES? ']'
  TAG      ::= '=' TAGID
  TYPE     ::= '(' LITERAL ')'
  FEATURES ::= FEATURE PATHSEP FSTRING FEATURES?

Match-rules for the tokens 'TAGID' and 'FEATURE'
are defined in terms of perl regular expressions, as follows:

  TAGID ::= /\w+/
  FEATURE ::= /[^:\s]*/

The math-rule for 'PATHSEP' tokens is determined by a
call to C<path_split_regex()>.

Whitespace is ignored, except in literal type-names, which may
contain nested parentheses.  Substructures sharing the same tag
are taken to be node-identical.  Returns the FStructure.

=item * C<as_text()>

Returns a formatted text string (as accepted by C<from_text()>)
representing the FStructure.

=item * C<as_string()>

Returns a text string (as accepted by C<from_string()>) representing
the FStructure.  Unlike the C<as_text()> method,
C<as_string()> does not insert newlines or indentation whitespace
into the strings it returns.

=item * C<as_pretty()>

Returns a pretty-printed string representation of the
FStructure.

=back

=cut


#--------------------------------------------------------------
# Other Viewing Methods
#--------------------------------------------------------------
=pod

=head2 Other Viewing Methods

=over 4

=item * C<as_graphviz()>

Returns a GraphViz object representing the FStructure.

See also: L<GraphViz>.

=item * C<as_latex_avm()>

Returns a string containing the LaTeX commands for representing the FStructure
with Chris Manning's `avm' package.

=item * C<viewps()>

Creates a temporary PostScript file containing the graph-representation
of the FStructure generated by C<as_graphviz()>, and calls
C<$PS_VIEWER> on this file.  Handy during debug sessions.

Temporary files are removed on normal program completion.

=item * C<viewdvi()>

Like C<viewps()>, but generates a temporary LaTeX file
with the C<as_latex_avm()> method, runs LaTeX on it,
and calls C<$DVI_VIEWER>.

Temporary files are removed on normal program completion.

=back

=cut


#--------------------------------------------------------------
# Low-level Utilties
#--------------------------------------------------------------
=pod

=head2 Low-level Utilties

=over 4

=item * C<as_tagged()>

Low-level method for generating "tagged" representations of
FStructures -- used by C<as_text()>, C<as_pretty()>, and
C<as_latex_avm()>.

=back

=cut


#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

File::Temp perl module by Tim Jenness.

Text::Balanced perl module by Damian Conway.

GraphViz perl module by Leon Brocard.

'avm' package for LaTeX by Chris Manning.


=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Data::FStructure(3pm).
Data::FStructure::perldata(3pm).
Math::PartialOrder(3pm).
Text::Balanced(3pm).
GraphViz(3pm).
latex(1).
