# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::theta.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Feature-Structures for perl: theta function
#
###############################################################

package Data::FStructure::theta;

our $VERSION = 0.02;

###############################################################
# Constructor
#   + Data::FStructure::theta->new()
#   + Object-structure:
#     { node1 => type1, ... }
#   + Args: NONE
###############################################################
sub new { return bless {}, ref($_[0]) || $_[0]; }

###############################################################
# Lookup
###############################################################

# $th->fetch($node);
sub fetch { return defined($_[1]) ? $_[0]->{$_[1]} : undef; }

# $val = $th->store($node,$val);
sub store {
  return
    (defined($_[1])
     ? $_[0]->{$_[1]} = $_[2]
     : undef);
}

# $bool = $th->is_eq($node,$val)
sub is_eq { return defined($_[1]) && $_[0] eq $_[1]; }


###############################################################
# Manipulation
###############################################################

# $th1 = $th1->merge($th2) : override
sub merge {
  @{$_[0]}{keys(%{$_[1]})} = values(%{$_[1]}) if (defined($_[1]));
  return $_[0];
}

# $th_dst = $th_dst->assign($th_src);
sub assign {
  %{$_[0]} = %{$_[1]} if (defined($_[1]));
  return $_[0];
}

# $th = $th->remove(@nodes);
sub remove {
  delete(@{$_[0]}{@_[1..$#_]});
  return $_[0];
}

# $th = $th->replace(%old2new);
sub replace {
  my ($th,%rename) = @_;
  @{$_[0]}{values(%rename)} = @{$_[0]}{keys(%rename)};
  delete(@{$_[0]}{keys(%rename)});
  return $_[0];
}

# $th1 = $th1->clear();
sub clear { %{$_[0]} = (); return $_[0]; }


1;
__END__

# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::theta -
Feature structures for perl -- partial type-mapping function theta.

=head1 SYNOPSIS

  use Data::FStructure::theta;

  #
  # Methods
  #
  $th = Data::FStructure::theta->new();   # new type-map

  $type = $th->fetch($node);             # get type for $node
  $type = $th->store($node,$type);       # store type for $node

  $th->remove(@nodes);                   # remove mappings from @nodes
  $th->replace($old=>$new,...);          # rename nodes

  $bool = $th1->is_eq($th2);             # test object-identity

  $th->clear();                          # drop all mappings
  $th_dst->merge($th_src);               # merge/override
  $th_dst->assign($th_src);              # assignment



=head1 REQUIREMENTS

Just perl.

=cut

#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

C<Data::FStructure::theta> objects provide type-mappings to
FStructure objects.  Currently,
C<Data::FStructure::theta> objects are implemented as simple perl hashes.

=cut

#--------------------------------------------------------------
# Variables
#--------------------------------------------------------------
#=pod
#
#=head1 PACKAGE VARIABLES
#
#=cut

#--------------------------------------------------------------
# Methods
#--------------------------------------------------------------
=pod

=head1 METHODS

=over 4

=item * C<new()>

Returns a new C<Data::FStructure::theta> object.

=item * C<fetch($node)>

Returns the type assigned to C<$node>, or C<undef>
if no type is currently assigned to C<$node>.

=item * C<store($node,$type)>

Assign the type C<$type> to the node C<$node>,
returns C<$type>.

=item * C<remove(@nodes)>

Removes type-assignments for all the nodes in
@nodes.  Returns the C<theta> object.

=item * C<< replace($old=E<gt>$new,...) >>

Assign types previously assigned to the C<$old> nodes
to the corresponding C<$new> nodes, and subsequently
remove the C<$old> nodes.  Returns the C<theta> object.

=item * C<clear()>

Removes all type-assignments.  Returns the C<theta> object.

=item * C<< $th1-E<gt>is_eq($th2) >>

Returns a true value if C<$th1> is object-identical
to C<$th2>, false otherwise.  Currently, this just checks
string-identity with C<eq>, but that should be pretty safe.

=item * C<< $th_dst-E<gt>merge($th_src) >>

Merges all type-assignments from C<$th_src> into C<$th_dst>,
overriding any assignments for shared node-names.  Returns
C<$th_dst>.

=item * C<< $th_dst-E<gt>assign($th_src) >>

Makes C<$th_dst> an object-distinct copy of C<$th_src>.

Note that types which are stored as references will
be shared.

=back

=cut


#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

Bob Carpenter, I<The Logic of Typed Feature Structures>.
Cambridge University Press, 1992.

perl(1).
Data::FStructure(3pm).
Data::FStructure::perldata(3pm).
Data::FStructure::theta(3pm).
Data::FStructure::delta(3pm).
Data::FStructure::pretty(3pm).
